//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// Cstat
#include <XSFit/StatMethod/Cstat/Cstat.h>
#include <XSFit/Fit/Fit.h>
#include <XSModel/Model/Model.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSUtil/Error/Error.h>
#include <XSstreams.h>
#include <XSContainer.h>
#include <set>
#include <typeinfo>
using namespace XSContainer;


// Class Cstat 
const Real Cstat::s_FLOOR = 1.0e-5;

Cstat::Cstat()
  : StatMethod("cstat"),
    m_firstDerivCoefficient(),
    m_secondDerivCoefficient(),
    m_fullName("C-Statistic"),
    m_scriptName("C-Statistic")
{
  isDerivAnalytic(true);
}

Cstat::Cstat(const Cstat &right)
  : StatMethod(right),
    m_firstDerivCoefficient(right.m_firstDerivCoefficient),
    m_secondDerivCoefficient(right.m_secondDerivCoefficient),
    m_fullName(right.m_fullName),
    m_scriptName(right.m_scriptName)
{
   if (!checkForPoisson())
      tcout << "Warning: Cstat statistic is only valid for Poisson data." << std::endl;
}


void Cstat::doPerform (Fit* fit)
{
    doReset(fit);

    Real cstat(0.);

    // for each model,
    ArrayContainer tmpQ(protoWorkspace());
    ModelMapConstIter ml (models->modelSet().begin());
    ModelMapConstIter mlEnd (models->modelSet().end());
    while (ml != mlEnd)
    {
        // add up the folded models for each spectrum [ >= indicates multiple sources]
        Model* m (ml->second);
        if (!m->isActive()) 
        {
           ++ml;
           continue;
        }
        ArrayContainer::const_iterator s = m->foldedModel().begin();
        ArrayContainer::const_iterator sEnd = m->foldedModel().end();
        while (s != sEnd)
        {
                size_t spectrumIndex (s->first);
                SpectralData* spectrum = m->spectrum(spectrumIndex);          
                const std::valarray<size_t>& IN = spectrum->indirectNotice();
                RealArray mod (s->second[IN]);
                if (tmpQ[spectrumIndex].size() == 0) tmpQ[spectrumIndex].resize(IN.size(),0.);
                tmpQ[spectrumIndex] += mod;
                ++s;
        }
        ++ml;
    }


    std::map<size_t,SpectralData*>::const_iterator   sp (fit->spectraToFit().begin());
    std::map<size_t,SpectralData*>::const_iterator   spEnd (fit->spectraToFit().end());

    // for each spectrum sum contributions.
    while(sp != spEnd) 
    {
	size_t spectrumIndex = sp->first;
	const SpectralData* spectrum = sp->second;
        const std::valarray<size_t>& IN = spectrum->indirectNotice();
        const size_t N (IN.size());

        RealArray ts (spectrum->exposureTime()*spectrum->areaScale()[IN]);
	RealArray S (ts * spectrum->spectrum()[IN]);
	const RealArray& model = tmpQ[spectrumIndex];
        const Background* background (spectrum->background());        

        if (spectrum->correction() && spectrum->correctionScale() != 0.0)
        {
           tcout << "Cstat: Correction data being ignored for spectrum "
             << spectrum->spectrumNumber() << std::endl;
        }

        if (!background)
        {
                for ( size_t i = 0; i < N; ++i)
                {
		   Real si = S[i];
		   Real tsi = ts[i];
		   Real yi = std::max(model[i], s_FLOOR/tsi);
		   cstat += tsi*yi;
                   if ( si > 0 ) cstat += si*(std::log(si)-std::log(tsi*yi)-1);
                }
        }
        else
        {
	   // case of a background spectrum in use
           Real bTime (background->data()->exposureTime());
           RealArray tb (background->data()->areaScale()[IN]);
           tb *= background->data()->backgroundScale()[IN];
           tb /= spectrum->backgroundScale()[IN];
           tb *= bTime;
           RealArray B (tb * background->spectrum()[IN]);

           // for notation, see the corresponding appendix in the XSPEC12 manual.
           for ( size_t i = 0; i < N; ++i)
           {
	      Real si = S[i];
	      Real bi = B[i];
	      Real tsi = ts[i];
	      Real tbi = tb[i];
              Real ti = ts[i] + tb[i];
	      Real yi = model[i];
              Real fi (0);

	      if ( fabs(si) <= SMALL )
	      {
		 cstat += tsi*yi - bi*std::log(tbi/ti);
	      }
	      else
	      {
		 if ( fabs(bi) <= SMALL )
		 {
		    if ( yi <= si/ti )
		    {
		       cstat += -tbi*yi - si*std::log(tsi/ti);
		    }
		    else
		    {
		       cstat += tsi*yi + si*(std::log(si)-std::log(tsi*yi)-1);
		    }
		 }
		 else
		 {
		    // now the main case where both data and background !=0
                    // Solve quadratic equation for fi, using Numerical Recipes technique
                    // to avoid round-off error that can easily cause problems here
                    // when b^2 >> ac.
                    Real a = ti;
                    Real b = ti*yi - si - bi;
                    Real c = -bi*yi;
                    Real sign = (b >= .0) ? 1.0 : -1.0;
                    Real q = -.5*(b + sign*sqrt(b*b - 4.0*a*c));
                    fi = q/a;
                    if (fi < .0) fi = c/q;

		    // note that at this point fi must be > 0 so the log 
		    // functions below will be valid.

                    cstat += tsi*yi + ti*fi - si*std::log(tsi*yi+tsi*fi)
		      - bi*std::log(tbi*fi) - si*(1-std::log(si)) 
		      - bi*(1-std::log(bi));

		 }

              }
           }  // end bin loop
	} // end if background

	++sp;

    } // end spectra loop

    Real bayesianPrior = bayes().lPrior(fit);           
    statistic(2.0*cstat - 2.0*bayesianPrior);
}

void Cstat::doInitialize (Fit* fit)
{

  StatMethod::doInitialize(fit);

  // allocate for derivative coefficient arrays.
  std::map<size_t,SpectralData*>::const_iterator i = fit->spectraToFit().begin();
  std::map<size_t,SpectralData*>::const_iterator iEnd = fit->spectraToFit().end();
  while (i != iEnd)
  {
        size_t N = i->second->indirectNotice().size();
        m_firstDerivCoefficient[i->first].resize(N,0.);
        m_secondDerivCoefficient[i->first].resize(N,0.);
        ++i;
  }
}

void Cstat::doDeallocate ()
{
    StatMethod::doDeallocate();

    m_firstDerivCoefficient.clear();
    m_secondDerivCoefficient.clear();
}

void Cstat::doReset (Fit* fit)
{

    std::set<string>::iterator ss (fit->activeModels().begin());
    std::set<string>::iterator ssEnd (fit->activeModels().end());
    while ( ss != ssEnd)
    {
          models->calculate(*ss);
          models->fold(*ss);
	  ++ss;
    }   

    // for each model,
    ArrayContainer tmpQ(protoWorkspace()); 
    ModelMapConstIter ml (models->modelSet().begin());      
    ModelMapConstIter mlEnd (models->modelSet().end());      
    while (ml != mlEnd)
    {
        // add up the folded models for each spectrum [ >= indicates multiple sources]
        Model* m (ml->second);
        if (!m->isActive()) 
        {
           ++ml;
           continue;
        }
        ArrayContainer::const_iterator s    = m->foldedModel().begin();
        ArrayContainer::const_iterator sEnd = m->foldedModel().end();
        while (s != sEnd)
        {
                size_t spectrumIndex (s->first);
                SpectralData* spectrum = m->spectrum(spectrumIndex);
                const std::valarray<size_t>& IN = spectrum->indirectNotice();
                RealArray mod (s->second[IN]);
                tmpQ[spectrumIndex] += mod;
                ++s;
        }
        ++ml;
    }    

    std::map<size_t,SpectralData*>::const_iterator sp    = fit->spectraToFit().begin();
    std::map<size_t,SpectralData*>::const_iterator spEnd = fit->spectraToFit().end();

    while(sp != spEnd) 
    {
        size_t spectrumIndex = sp->first;
        const SpectralData* spectrum  = sp->second;        
        const std::valarray<size_t>& IN  = spectrum->indirectNotice();
        const size_t N (IN.size());

        RealArray ts (spectrum->exposureTime()*spectrum->areaScale()[IN]);
        RealArray S (ts * spectrum->spectrum()[IN]);
        RealArray& diff1 = m_firstDerivCoefficient[spectrumIndex];
        RealArray& diff2 = m_secondDerivCoefficient[spectrumIndex];
        RealArray& model = tmpQ[spectrumIndex];
        const Background* background (spectrum->background());
        if (!background)
        {
           for ( size_t i = 0; i < N; ++i)
           {
 	      Real si = S[i];
	      Real tsi = ts[i];
              Real yi = std::max(model[i], s_FLOOR/tsi);
              diff1[i] = (si/yi) - tsi;
              diff2[i] = si/(yi*yi);
           }
        }
        else
        {
           // Variable definitions follow Appendix B of manual.
           Real bTime (background->data()->exposureTime());
           RealArray tb (background->data()->areaScale()[IN]);
           tb *= background->data()->backgroundScale()[IN];
           tb /= spectrum->backgroundScale()[IN];
           tb *= bTime;
           RealArray B (tb * background->spectrum()[IN]);

           for (size_t i=0; i<N; ++i)
           {
              Real si = S[i];
              Real bi = B[i];
	      Real tsi = ts[i];
	      Real tbi = tb[i];
	      Real ti = ts[i] + tb[i];
              Real yi = std::max(model[i], s_FLOOR/tsi);
	      if (fabs(si) <= SMALL)
	      {
		 // Special case if the data = 0 - note that in this case the 
		 // derivatives are independent of the background values
		 diff1[i] = -tsi;
		 diff2[i] = 0.;
	      }
              else
              {
                 if (fabs(bi) <= SMALL)
                 {
                    // If background bin is 0 but data is non-zero
		    if (yi <= si/ti )
		    {
		       diff1[i] = tbi;
		       diff2[i] = 0.;
		    }
		    else
		    {
 		       diff1[i] = (si/yi) - tsi;
		       diff2[i] = si/(yi*yi);
                    }
		 }
		 else
		 {
                    // Solve quadratic equation for fi, using Numerical Recipes technique
                    // to avoid round-off error that can easily cause problems here
                    // when b^2 >> ac.
                    Real a = ti;
		    Real b = ti*yi - si - bi;
		    Real c = -bi*yi;
		    Real sign = (b >= .0) ? 1.0 : -1.0;
                    Real q = -.5*(b + sign*sqrt(b*b - 4.0*a*c));
                    Real fi = q/a;
		    Real gi = 0.;
		    Real hi = 0.;
		    Real di = sqrt(b*b - 4.0*a*c);
                    if (fi < .0)
                    {
		       fi = c/q;
                       gi = (ti*yi - si + bi - di)/(2.0*di);
                       hi = 2.0*ti*si*bi/(di*di*di);
                    }
                    else
                    {
                       gi = (-ti*yi + si - bi - di)/(2.0*di);
                       hi = -2.0*ti*si*bi/(di*di*di);
                    }
                    diff1[i] = si*(1+gi)/(yi+fi) + bi*gi/fi - ts[i] -gi*ti;
                    diff2[i] = si*hi/(yi+fi) + si*(1+gi)*(1+gi)/((yi+fi)*(yi+fi)) 
                                 + bi*hi/fi + bi*gi*gi/(fi*fi) - ti*hi;
                 }
	      }

	   }
        }
	++sp;
    }
}

Cstat* Cstat::clone () const
{
  return new Cstat(*this);
}

void Cstat::doReport (Fit* fit) const
{
  int degrees (0);
  size_t bins(0);
  degreesOfFreedom(fit,degrees,bins);
  using namespace std;
  ios_base::fmtflags saveFmt(tcout.flags());
  streamsize savePrec(tcout.precision());
  Real truncatedStat = setOutputFormat(fit);
  tcout << "\n C-statistic = " << setw(14) << truncatedStat << " using " 
                  << bins << " PHA bins and " << degrees 
                  << " degrees of freedom." << endl;
  if (!checkForPoisson())
      tcout << "\nWarning: Cstat statistic is only valid for Poisson data.\n" << std::endl;
  tcout.precision(savePrec);
  tcout.flags(saveFmt);
}

Real Cstat::plotChi (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{
  Real point (0);
  // Presumeably areaTime is not .0 or it wouldn't have made it here.
  if ( obs > 0 )
  {
       obs *= areaTime;
       mod *= areaTime;
       if (mod > 0)
       {
          Real sign = (obs >= mod) ? 1.0 : -1.0;
          point = 2.0*sign*(mod - obs*(1 - (log(obs) - log(mod))));
       }
       else
          point = NODATA;
  }
  else
  {
        point = 2*mod*areaTime;       
  }      
  return point;                    
}

ArrayContainer Cstat::getDifferences (const Fit* fit) const
{
  using namespace XSContainer;
  ArrayContainer result;

  SpectralDataMapConstIt  sm (fit->spectraToFit().begin());  
  SpectralDataMapConstIt  smEnd (fit->spectraToFit().end());  

  while (sm != smEnd)
  {
        size_t index = sm->first;
        const SpectralData* spectrum = sm->second;
        const std::valarray<size_t>& IN = spectrum->indirectNotice();
        result[index].resize(IN.size());
        result[index] = spectrum->spectrum()[IN];
        if (spectrum->background())
        {
                result[index] -= spectrum->background()->spectrum()[IN];
                if (spectrum->correction() && spectrum->correctionScale() != 0) 
                {
                        result[index] -=  spectrum->correctionScale()
                                *spectrum->correction()->spectrum()[IN];         
                }     
        }
        ++sm;  
  }


  ModelMapConstIter ml (models->modelSet().begin());      
  ModelMapConstIter  mlEnd (models->modelSet().end()); 
  while ( ml != mlEnd )
  {
        // D - F for each model. In the case where there are multiple
        // sources difference will be iteratively defined.
        Model* m (ml->second);
        if (m->isActive())
        {
                m->difference(result);                  
        }
        ++ml;
  }    
  return result;
}

Real Cstat::sumDerivs (int parameterIndex, bool isRespar)
{

    //dModel = container of array containers of derivatives for each parameter.
    ArrayContainer& dm = dModel()[parameterIndex];
    Real dSdpj(0);

    ArrayContainer::iterator d = dm.begin();
    ArrayContainer::iterator dEnd = dm.end();
    if (!isRespar)
    {
       while(d != dEnd) 
       {
          size_t spectrumIndex = d->first;
          const RealArray& dv = d->second;
          RealArray dSdpA (m_firstDerivCoefficient[spectrumIndex]);
          dSdpA *= dv;
          dSdpj += -2.0*dSdpA.sum();
          if (false)
          {
	      size_t N = m_firstDerivCoefficient[spectrumIndex].size();
              tcout << N << std::endl;
	      std::vector<Real> PS(N);
	      std::partial_sum(&dSdpA[0],&dSdpA[N],PS.begin());
	      tcout << " Spectrum " << d->first << " Parameter# " << parameterIndex  
                              << " D:  " << dSdpj << '\n';
	      for (size_t j = 0; j < N; ++j)
	      {
	          tcout << j << " " << std::setw(12) 
		        << std::setw(12) << m_firstDerivCoefficient[spectrumIndex][j] << " "
		        << std::setw(12) << dv[j]  << " "
		        << std::setw(12) << dSdpA[j] << " "
		        << std::setw(12) << PS[j]  << '\n'; 
	      }
	      tcout << " d " << spectrumIndex << "  " << dSdpj << '\n';  
          }               
          ++d;
       }
    }
    else
    {
       std::map<size_t,ArrayContainer>::const_iterator dFdPContainer = dRM().find(parameterIndex);
       // For response parameters, dFdPContainer should only have
       // 1 entry, since only 1 spectrum is associated with reponse.
       int specNum = dFdPContainer->second.begin()->first;
       const RealArray& dFdP = dFdPContainer->second.begin()->second;
       RealArray dSdpA(m_firstDerivCoefficient[specNum]);
       dSdpA *= dFdP;
       dSdpj += -2.0*dSdpA.sum();
    }
    return dSdpj;
}

Real Cstat::sumSecondDerivs (int pi, int pj, bool isPiResp, bool isPjResp)
{
  Real alpha_ij=.0;
  ArrayContainer::const_iterator secondCoeffEnd = m_secondDerivCoefficient.end();
  if (!isPiResp && !isPjResp)
  {
     // Calculate (2/var)*(dF/dpi)(dF/dpj) summed over all fit spectra.
     const ArrayContainer& piBins = dModel().find(pi)->second;
     const ArrayContainer& pjBins = dModel().find(pj)->second;
     ArrayContainer::const_iterator piSpec = piBins.begin();
     ArrayContainer::const_iterator piEnd = piBins.end();
     ArrayContainer::const_iterator pjEnd = pjBins.end();
     while (piSpec != piEnd)
     {
        // ASSUMES piBins and pjBins containers have same number of
        // RealArrays, which are indexed by the same spectra numbers.
        // Does NOT ASSUME that iterating through both in parallel will
        // grab RealArrays corresponding to the same spectrum.
        size_t specNum = piSpec->first;
        ArrayContainer::const_iterator pjSpec = pjBins.find(specNum);
        ArrayContainer::const_iterator coeffSpec = m_secondDerivCoefficient.find(specNum);  
        if (pjSpec == pjEnd || coeffSpec == secondCoeffEnd)
        {
           throw RedAlert("Cstat calc: Second derivatives array container size mismatch.");
        }
        RealArray secondDeriv(coeffSpec->second);
        secondDeriv *= piSpec->second;
        secondDeriv *= pjSpec->second;
        alpha_ij += secondDeriv.sum();
        ++piSpec;
     } 
  }
  else
  {
     // 1 or both pars are response pars.   In all 3 possible
     // cases, only 1 spectrum (at most) applies.
     size_t onlySpec = 0;
     std::map<size_t,ArrayContainer>::const_iterator piContainer = 
                dModel().find(pi);
     std::map<size_t,ArrayContainer>::const_iterator pjContainer = 
                dModel().find(pj);
     if (isPiResp)
     {
        piContainer = dRM().find(pi);
        if (piContainer == dRM().end())
        {
           throw RedAlert("Cstat 2nd deriv: missing response parameter");
        }
        onlySpec = piContainer->second.begin()->first;
     }
     if (isPjResp)
     {
        pjContainer = dRM().find(pj);
        if (pjContainer == dRM().end())
        {
           throw RedAlert("Cstat 2nd deriv: missing response parameter");
        }
        size_t tmp = pjContainer->second.begin()->first;
        if (onlySpec && (tmp != onlySpec))
        {
           // 2 response pars belonging to different spectra.
           // No terms contribute to 2nd derivative.
           return 0.0;
        }
        onlySpec = tmp;
     }
     const RealArray& piSpec = piContainer->second.find(onlySpec)->second;
     const RealArray& pjSpec = pjContainer->second.find(onlySpec)->second;
     ArrayContainer::const_iterator coeffSpec = m_secondDerivCoefficient.find(onlySpec);  
     if (coeffSpec == secondCoeffEnd)
     {
        throw RedAlert("Cstat calc: Second derivatives array container size mismatch.");
     }
     RealArray secondDeriv(coeffSpec->second);
     secondDeriv *= piSpec;
     secondDeriv *= pjSpec;
     alpha_ij = secondDeriv.sum();
  }
  alpha_ij *= 2.0;  
  return alpha_ij;
}

std::pair<Real,Real> Cstat::adjustForFrozen (const SpectralData* spec, const std::vector<Model*>& models, const std::map<Model*,ArrayContainer >& savedValsForFrozen)
{

    const std::valarray<size_t>& IN = spec->indirectNotice();
    const size_t N = IN.size();
    const size_t specNum = spec->spectrumNumber();

    std::pair<Real,Real> fSums (0.,0.);

    RealArray data (spec->spectrum()[IN]);
    RealArray totModFroz(0., N); 
    RealArray totalModel(0., N);

    for (size_t i=0; i<models.size(); ++i)
    {
      Model* mod = models[i];
      std::map<Model*,ArrayContainer >::const_iterator itSaved =
                savedValsForFrozen.find(mod);
      if (itSaved != savedValsForFrozen.end())
      {
         totalModel += itSaved->second.find(specNum)->second[IN];
         totModFroz += mod->foldedModel(specNum)[IN];
      }
      else
      {
         totalModel += mod->foldedModel(specNum)[IN];
      }
    }

    fSums.first = (data - totModFroz).sum();
    fSums.second = (totalModel - totModFroz).sum();

   return fSums;
}

bool Cstat::checkForPoisson ()
{
   bool allPoisson = true;
   std::map<size_t,SpectralData*>::const_iterator itSpec = 
                fit->spectraToFit().begin();
   std::map<size_t,SpectralData*>::const_iterator itSpecEnd = 
                fit->spectraToFit().end();
   while (itSpec != itSpecEnd)
   {
      const SpectralData* sd = itSpec->second;
      if (!sd->isPoisson())
      {
         allPoisson = false;
         break;
      }
      if (sd->background())
      {
         if (!sd->background()->data()->isPoisson())
         {
            allPoisson = false;
            break;
         }
      }
      ++itSpec;
   }
   return allPoisson;
}

// Additional Declarations
