//   Read the documentation to learn more about C++ code generator
//   versioning.
//	  %X% %Q% %Z% %W%

// Lstat
#include <XSFit/StatMethod/Lstat/Lstat.h>
#include <XSContainer.h>
#include <XSFit/Fit/Fit.h>
#include <XSFit/StatMethod/Lstat/LstatT.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSModel/Model/Model.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSUtil/Error/Error.h>
#include <XSUtil/Numerics/LnBang.h>
#include <XSstreams.h>
#include <cmath>
#include <set>
#include <typeinfo>
using namespace XSContainer;


// Class Lstat::Dterm0 
const int Lstat::Dterm0::flag = 0;

// Class Lstat::Dterm1 
const int Lstat::Dterm1::flag = 1;

// Class Lstat::Dterm2 
const int Lstat::Dterm2::flag = 2;

// Class Lstat 
const Real Lstat::s_FLOOR = 1.0e-5;

Lstat::Lstat()
  : StatMethod("lstat"),
    m_firstDerivCoefficient(),
    m_secondDerivCoefficient(),
    m_sourceCounts(0),
    m_backgroundCounts(0),
    m_fullName("L-statistic"),
    m_scriptName("L-statistic")
{
  isDerivAnalytic(true);
}

Lstat::Lstat(const Lstat &right)
  : StatMethod(right),
    m_firstDerivCoefficient(right.m_firstDerivCoefficient),
    m_secondDerivCoefficient(right.m_secondDerivCoefficient),
    m_sourceCounts(right.m_sourceCounts),
    m_backgroundCounts(right.m_backgroundCounts),
    m_fullName(right.m_fullName),
    m_scriptName(right.m_scriptName)
{
   if (!checkForPoisson())
      tcout << "Warning: Lstat statistic is only valid for Poisson data." << std::endl;
}


void Lstat::doPerform (Fit* fit)
{

   doReset(fit);

   ArrayContainer tmp;
   ModelMapConstIter ml (models->modelSet().begin());
   ModelMapConstIter mlEnd (models->modelSet().end());

   // for each model,
   while (ml != mlEnd)
   {
       // add up the folded models for each spectrum [ >= indicates multiple sources]
       Model* m (ml->second);
       if (!m->isActive()) 
       {
          ++ml;
          continue;
       }
       ArrayContainer::const_iterator s = m->foldedModel().begin();
       ArrayContainer::const_iterator sEnd = m->foldedModel().end();
       while (s != sEnd)
       {
               SpectralData* obs = m->spectrum(s->first);          
               const std::valarray<size_t>& IN = obs->indirectNotice();
               size_t spectrumIndex = obs->spectrumNumber();
               RealArray mod (s->second[IN]);
               if (tmp[spectrumIndex].size() == 0) tmp[spectrumIndex].resize(IN.size(),0.);
               tmp[spectrumIndex] += mod;
               ++s;
       }
       ++ml;
   }

   std::map<size_t,SpectralData*>::const_iterator   sm (fit->spectraToFit().begin());
   std::map<size_t,SpectralData*>::const_iterator   smEnd (fit->spectraToFit().end());

   // for each spectrum sum contributions.
   Real lstat = .0;
   // for each spectrum sum contributions.
   while(sm != smEnd) 
   {
       size_t index = sm->first;
       const SpectralData* spectrum = sm->second;

       const std::valarray<size_t>& IN = spectrum->indirectNotice();
       const size_t N (IN.size());

       RealArray ts (spectrum->exposureTime()*spectrum->areaScale()[IN]);
       RealArray modelSum(0.,N);

       if (spectrum->correction() && spectrum->correctionScale() != 0.0)
       {
          tcout << "Lstat: Correction data being ignored for spectrum "
            << spectrum->spectrumNumber() << std::endl;
       }
       const Background* background (spectrum->background());       
       if (!background)
       {
          // same as cstat
          const RealArray& model = tmp[index];
          RealArray mod  ( ts*model );
          RealArray dat ( ts*spectrum->spectrum()[IN] );
          for ( size_t i = 0; i < N; ++i)
          {
             Real lmod = std::log(std::max(s_FLOOR,mod[i]));
             Real ldat = std::log(std::max(s_FLOOR,dat[i]));
             modelSum[i] = mod[i] - dat[i] + dat[i]*(ldat - lmod);
          }          
       }
       else
       {
          Real bTime (background->data()->exposureTime());
          RealArray tb (background->data()->areaScale()[IN]);
          tb *= background->data()->backgroundScale()[IN];
          tb /= spectrum->backgroundScale()[IN];
          tb *= bTime;
          RealArray S (ts * spectrum->spectrum()[IN]);
          RealArray B (tb * background->spectrum()[IN]);
          RealArray saturated(spectrum->spectrum()[IN] - background->spectrum()[IN]);
          const RealArray& y = tmp[index];
          for (size_t i=0; i<N; ++i)
          {
             m_sourceCounts = static_cast<int>(std::floor(S[i]+.5));
             m_backgroundCounts = static_cast<int>(std::floor(B[i]+.5));
             Real totScale = ts[i] + tb[i];
             Real mod = y[i];
             Real sat = std::max(.0, saturated[i]);

             // The lorsum routine returns the log of the summation 
             // from 0 to nobs.
             Real term1 = lorSum<Dterm0>(mod*totScale);
             modelSum[i] = mod*ts[i] - term1;
             Real term2 = lorSum<Dterm0>(sat*totScale);
             modelSum[i] += term2 - sat*ts[i];
          }  
       }
       lstat += modelSum.sum();
       ++sm;
   }

   Real bayesianPrior = bayes().lPrior(fit);           
   statistic(2.0*lstat - 2.0*bayesianPrior);
}

void Lstat::doInitialize (Fit* fit)
{

  StatMethod::doInitialize(fit);

  // allocate for derivative coefficient arrays.
  std::map<size_t,SpectralData*>::const_iterator i = fit->spectraToFit().begin();
  std::map<size_t,SpectralData*>::const_iterator iEnd = fit->spectraToFit().end();
  while (i != iEnd)
  {
        size_t N = i->second->indirectNotice().size();
        m_firstDerivCoefficient[i->first].resize(N,0.);
        m_secondDerivCoefficient[i->first].resize(N,0.);
        ++i;
  }
}

void Lstat::doDeallocate ()
{
  StatMethod::doDeallocate();
  m_firstDerivCoefficient.clear();
  m_secondDerivCoefficient.clear();
}

void Lstat::doReset (Fit* fit)
{

    std::set<string>::iterator ss (fit->activeModels().begin());
    std::set<string>::iterator ssEnd (fit->activeModels().end());
    while ( ss != ssEnd)
    {
          models->calculate(*ss);
          models->fold(*ss);
	  ++ss;
    }   

    // for each model,
    ArrayContainer tmpQ(protoWorkspace()); 
    ModelMapConstIter ml (models->modelSet().begin());      
    ModelMapConstIter mlEnd (models->modelSet().end());      
    while (ml != mlEnd)
    {
        // add up the folded models for each spectrum [ >= indicates multiple sources]
        Model* m (ml->second);
        if (!m->isActive()) 
        {
           ++ml;
           continue;
        }
        ArrayContainer::const_iterator s    = m->foldedModel().begin();
        ArrayContainer::const_iterator sEnd = m->foldedModel().end();
        while (s != sEnd)
        {
                size_t spectrumIndex (s->first);
                SpectralData* spectrum = m->spectrum(spectrumIndex);
                const std::valarray<size_t>& IN = spectrum->indirectNotice();
                RealArray mod (s->second[IN]);
                tmpQ[spectrumIndex] += mod;
                ++s;
        }
        ++ml;
    }    

    std::map<size_t,SpectralData*>::const_iterator sp    = fit->spectraToFit().begin();
    std::map<size_t,SpectralData*>::const_iterator spEnd = fit->spectraToFit().end();

    while(sp != spEnd) 
    {
        size_t spectrumIndex = sp->first;
        const SpectralData* spectrum  = sp->second;        
        const std::valarray<size_t>& IN  = spectrum->indirectNotice();
        const size_t N (IN.size());
        Real exposureTime (spectrum->exposureTime());
        RealArray area (spectrum->areaScale()[IN]);
        RealArray data (sp->second->spectrum()[IN]);
        RealArray& diff1 = m_firstDerivCoefficient[spectrumIndex];
        RealArray& diff2 = m_secondDerivCoefficient[spectrumIndex];
        RealArray& model = tmpQ[spectrumIndex];
        const Background* background (spectrum->background());
        if (!background)
        {
           for ( size_t j = 0; j < N; ++j)
           {
              Real atime = exposureTime*area[j];
              Real mod_j = std::max(model[j], s_FLOOR/atime);
              diff1[j] = atime*(data[j]/mod_j - 1.);
              diff2[j] = atime*data[j]/(mod_j*mod_j);
           }
        }
        else
        {
           // Variable definitions follow Appendix B of manual.
           RealArray ts (exposureTime*area);
           Real bTime (background->data()->exposureTime());
           RealArray tb (background->data()->areaScale()[IN]);
           tb *= background->data()->backgroundScale()[IN];
           tb /= spectrum->backgroundScale()[IN];
           tb *= bTime;
           RealArray S (ts * spectrum->spectrum()[IN]);
           RealArray B (tb * background->spectrum()[IN]);
	   RealArray totalScale(ts + tb);
           for (size_t i=0; i<N; ++i)
           {
	      Real t1 = std::log(totalScale[i]);
	      Real t2 = 2.0*t1;
	      Real totScale_i = totalScale[i];
	      Real mod_i = model[i];
              m_sourceCounts = static_cast<int>(std::floor(S[i]+.5));
              m_backgroundCounts = static_cast<int>(std::floor(B[i]+.5));
	      Real f = lorSum<Dterm0>(totScale_i*mod_i);
	      Real g = lorSum<Dterm1>(totScale_i*mod_i) + t1;
	      Real h = lorSum<Dterm2>(totScale_i*mod_i) + t2;

	      diff1[i] = std::exp(g - f) - ts[i];
	      diff2[i] = std::exp(2.0*(g - f)) - std::exp(h - f);
           }
        }
	++sp;
    }
}

Lstat* Lstat::clone () const
{
  return new Lstat(*this);   
}

void Lstat::doReport (Fit* fit) const
{
  int degrees (0);
  size_t bins(0);
  degreesOfFreedom(fit,degrees,bins);
  using namespace std;
  ios_base::fmtflags saveFmt(tcout.flags());
  streamsize savePrec(tcout.precision());
  Real truncatedStat = setOutputFormat(fit);  
  tcout << "\n L-statistic = " << setw(14) << truncatedStat << " using " 
                  << bins << " PHA bins and " << degrees 
                  << " degrees of freedom." << endl;
  if (!checkForPoisson())
     tcout << "\nWarning: Lstat statistic is only valid for Poisson data.\n" << std::endl;
  tcout.precision(savePrec);
  tcout.flags(saveFmt);
}

Real Lstat::sumDerivs (int parameterIndex, bool isRespar)
{

    //dModel = container of array containers of derivatives for each parameter.
    ArrayContainer& dm = dModel()[parameterIndex];
    Real dSdpj(0);

    ArrayContainer::iterator d = dm.begin();
    ArrayContainer::iterator dEnd = dm.end();
    if (!isRespar)
    {
       while(d != dEnd) 
       {
          size_t spectrumIndex = d->first;
          const RealArray& dv = d->second;
          RealArray dSdpA (m_firstDerivCoefficient[spectrumIndex]);
          dSdpA *= dv;
          dSdpj += -2.0*dSdpA.sum();
          ++d;
       }
    }
    else
    {
       std::map<size_t,ArrayContainer>::const_iterator dFdPContainer = dRM().find(parameterIndex);
       // For response parameters, dFdPContainer should only have
       // 1 entry, since only 1 spectrum is associated with reponse.
       int specNum = dFdPContainer->second.begin()->first;
       const RealArray& dFdP = dFdPContainer->second.begin()->second;
       RealArray dSdpA(m_firstDerivCoefficient[specNum]);
       dSdpA *= dFdP;
       dSdpj += -2.0*dSdpA.sum();
    }
    return dSdpj;
}

Real Lstat::sumSecondDerivs (int pi, int pj, bool isPiResp, bool isPjResp)
{
  Real alpha_ij=.0;
  ArrayContainer::const_iterator secondCoeffEnd = m_secondDerivCoefficient.end();
  if (!isPiResp && !isPjResp)
  {
     // Calculate (2/var)*(dF/dpi)(dF/dpj) summed over all fit spectra.
     const ArrayContainer& piBins = dModel().find(pi)->second;
     const ArrayContainer& pjBins = dModel().find(pj)->second;
     ArrayContainer::const_iterator piSpec = piBins.begin();
     ArrayContainer::const_iterator piEnd = piBins.end();
     ArrayContainer::const_iterator pjEnd = pjBins.end();
     while (piSpec != piEnd)
     {
        // ASSUMES piBins and pjBins containers have same number of
        // RealArrays, which are indexed by the same spectra numbers.
        // Does NOT ASSUME that iterating through both in parallel will
        // grab RealArrays corresponding to the same spectrum.
        size_t specNum = piSpec->first;
        ArrayContainer::const_iterator pjSpec = pjBins.find(specNum);
        ArrayContainer::const_iterator coeffSpec = m_secondDerivCoefficient.find(specNum);  
        if (pjSpec == pjEnd || coeffSpec == secondCoeffEnd)
        {
           throw RedAlert("Lstat calc: Second derivatives array container size mismatch.");
        }
        RealArray secondDeriv(coeffSpec->second);
        secondDeriv *= piSpec->second;
        secondDeriv *= pjSpec->second;
        alpha_ij += secondDeriv.sum();
        ++piSpec;
     } 
  }
  else
  {
     // 1 or both pars are response pars.   In all 3 possible
     // cases, only 1 spectrum (at most) applies.
     size_t onlySpec = 0;
     std::map<size_t,ArrayContainer>::const_iterator piContainer = 
                dModel().find(pi);
     std::map<size_t,ArrayContainer>::const_iterator pjContainer = 
                dModel().find(pj);
     if (isPiResp)
     {
        piContainer = dRM().find(pi);
        if (piContainer == dRM().end())
        {
           throw RedAlert("Lstat 2nd deriv: missing response parameter");
        }
        onlySpec = piContainer->second.begin()->first;
     }
     if (isPjResp)
     {
        pjContainer = dRM().find(pj);
        if (pjContainer == dRM().end())
        {
           throw RedAlert("Lstat 2nd deriv: missing response parameter");
        }
        size_t tmp = pjContainer->second.begin()->first;
        if (onlySpec && (tmp != onlySpec))
        {
           // 2 response pars belonging to different spectra.
           // No terms contribute to 2nd derivative.
           return 0.0;
        }
        onlySpec = tmp;
     }
     const RealArray& piSpec = piContainer->second.find(onlySpec)->second;
     const RealArray& pjSpec = pjContainer->second.find(onlySpec)->second;
     ArrayContainer::const_iterator coeffSpec = m_secondDerivCoefficient.find(onlySpec);  
     if (coeffSpec == secondCoeffEnd)
     {
        throw RedAlert("Lstat calc: Second derivatives array container size mismatch.");
     }
     RealArray secondDeriv(coeffSpec->second);
     secondDeriv *= piSpec;
     secondDeriv *= pjSpec;
     alpha_ij = secondDeriv.sum();
  }
  alpha_ij *= 2.0;  
  return alpha_ij;
}

ArrayContainer Lstat::getDifferences (const Fit* fit) const
{
   ArrayContainer result;

  SpectralDataMapConstIt  sm (fit->spectraToFit().begin());  
  SpectralDataMapConstIt  smEnd (fit->spectraToFit().end());  

  while (sm != smEnd)
  {
        size_t index = sm->first;
        const SpectralData* spectrum = sm->second;
        const std::valarray<size_t>& IN = spectrum->indirectNotice();
        result[index].resize(IN.size());
        result[index] = spectrum->spectrum()[IN];
        if (spectrum->background())
        {
                result[index] -= spectrum->background()->spectrum()[IN];
                if (spectrum->correction() && spectrum->correctionScale() != 0) 
                {
                        result[index] -=  spectrum->correctionScale()
                                *spectrum->correction()->spectrum()[IN];         
                }     
        }
        ++sm;  
  }


  ModelMapConstIter ml (models->modelSet().begin());      
  ModelMapConstIter  mlEnd (models->modelSet().end()); 
  while ( ml != mlEnd )
  {
        // D - F for each model. In the case where there are multiple
        // sources difference will be iteratively defined.
        Model* m (ml->second);
        if (m->isActive())
        {
                m->difference(result);                  
        }
        ++ml;
  }    
   return result;
}

Real Lstat::plotChi (Real obs, Real mod, Real error, Real areaTime, const Real& NODATA) const
{
  Real point (0);
  // Presumeably areaTime is not .0 or it wouldn't have made it here.
  if ( obs > 0 )
  {
       obs *= areaTime;
       mod *= areaTime;
       if (mod > 0)
       {
          Real sign = (obs >= mod) ? 1.0 : -1.0;
          point = 2.0*sign*(mod - obs*(1 - (log(obs) - log(mod))));
       }
       else
          point = NODATA;
  }
  else
  {
        point = 2*mod*areaTime;       
  }      
  return point;                    
}

Real Lstat::lorTerm (const int n1, const int n2, const int n3)
{
   // Function used in the calculation of the 
   // background-marginalized likelihood.  
   // Returns  LOG ( (n1+n2-n3)!/(n1-n3)/!n3! ).
   Numerics::LnBang lnBang;
   Real result = lnBang(n1+n2-n3) - lnBang(n1-n3) - lnBang(n3);
   return result;
}

std::pair<Real,Real> Lstat::adjustForFrozen (const SpectralData* spec, const std::vector<Model*>& models, const std::map<Model*,ArrayContainer >& savedValsForFrozen)
{

    const std::valarray<size_t>& IN = spec->indirectNotice();
    const size_t N = IN.size();
    const size_t specNum = spec->spectrumNumber();

    std::pair<Real,Real> fSums (0.,0.);

    RealArray data (spec->spectrum()[IN]);
    RealArray totModFroz(0., N); 
    RealArray totalModel(0., N);

    for (size_t i=0; i<models.size(); ++i)
    {
      Model* mod = models[i];
      std::map<Model*,ArrayContainer >::const_iterator itSaved =
                savedValsForFrozen.find(mod);
      if (itSaved != savedValsForFrozen.end())
      {
         totalModel += itSaved->second.find(specNum)->second[IN];
         totModFroz += mod->foldedModel(specNum)[IN];
      }
      else
      {
         totalModel += mod->foldedModel(specNum)[IN];
      }
    }

    fSums.first = (data - totModFroz).sum();
    fSums.second = (totalModel - totModFroz).sum();

   return fSums;
}

bool Lstat::checkForPoisson ()
{
   bool allPoisson = true;
   std::map<size_t,SpectralData*>::const_iterator itSpec = 
                fit->spectraToFit().begin();
   std::map<size_t,SpectralData*>::const_iterator itSpecEnd = 
                fit->spectraToFit().end();
   while (itSpec != itSpecEnd)
   {
      const SpectralData* sd = itSpec->second;
      if (!sd->isPoisson())
      {
         allPoisson = false;
         break;
      }
      if (sd->background())
      {
         if (!sd->background()->data()->isPoisson())
         {
            allPoisson = false;
            break;
         }
      }
      ++itSpec;
   }
   return allPoisson;
}

// Additional Declarations
