//
//  XSPEC12  November 2003
//
//

// location for functions that are in namespace XSGlobal but not part
// of the Global class.

#include <XSUser/Global/Global.h>
#include <XSUser/Help/HelpComposite.h>
#include <XSUser/UserInterface/TclRedAlert.h>
#include <XSPlot/Plot/Plot.h>
#include <XSPlot/Plot/Graph.h>
#include <XSPlot/Plot/PlotCommand.h>
#include <XSPlot/Plot/PlotCommandCreator.h>
#include <XSFit/Fit/Fit.h>
#include <XSFit/Fit/FitMethod.h>
#include <XSFit/FitMethod/LevMarq/LevMarq.h>
#include <XSFit/MCMC/ChainManager.h>
#include <XSFit/Randomizer/Randomizer.h>
#include <XSFit/Randomizer/RandomizerPolicies.h>
#include <XSFit/StatMethod/ChiSquare/ChiSquare.h>
#include <XSFit/StatMethod/Cstat/Cstat.h>
#include <XSFit/StatMethod/EDF/EDF.h>
#include <XSFit/StatMethod/EDF/EDFVariants.h>
#include <XSFit/StatMethod/Lstat/Lstat.h>
#include <XSModel/GlobalContainer/ModelContainer.h>
#include <XSModel/GlobalContainer/ResponseContainer.h>
#include <XSModel/GlobalContainer/DataContainer.h>
#include <XSModel/GlobalContainer/Weight.h>
#include <XSModel/Model/Model.h>
#include <XSModel/Model/Component/Component.h>
#include <XSModel/Model/Component/TableComponent.h>
#include <XSModel/Model/Component/OGIPTable/OGIPTable.h>
#include <XSModel/Model/EmissionLines/LineList.h>
#include <XSModel/Model/EmissionLines/Apec.h>
#include <XSModel/Model/EmissionLines/Bearden.h>
#include <XSModel/Model/EmissionLines/Mekal.h>
#include <XSModel/Parameter/Parameter.h>
#include <XSModel/Data/DataSet.h>
#include <XSModel/Data/SpectralData.h>
#include <XSModel/Data/BackCorr/Background.h>
#include <XSUtil/Error/Error.h>
#include <XSUtil/Parse/XSparse.h>
#include <XSUtil/Signals/SignalHandler.h>
#include <XSUtil/Utils/XSutility.h>
#include <XSFunctions/FunctionUtility.h>
#include <XSFunctions/functionMap.h>
#include <Handler/HandlerUtils.h>
#include <XSGlobal.h>
#include <XSContainer.h>
#include <XSstreams.h>
#include <iostream>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>    
#include <cassert>
#include <memory>
#include <cstring>

extern "C" {
#ifdef GUI_SUPPORT
  #include <pow.h>
  #include <itk.h>
#endif
}

namespace xstcl
{
   // container definitions for the command set and their
   // one-line help summaries
   char xsLogOut[] = "LogOut";
   char xsLogErr[] = "ErrOut";
   int  xs_echo_script = 0;
   char xsTclPrompt[] = "xs_tcl_prompt";
   char* xs_tcl_prompt = 0;
   Tcl_Trace XS_Echo_Trace = 0;
}

namespace XSGlobal
{
    GlobalData* globalData = 0;
    std::map <string,Tcl_ObjCmdProc*> commandMap;
    std::map <string,string> summaryMap;
    std::map <string,string> pdfMap;        //good place for this?
}

bool
XSGlobal::printDocs(const char* command, const char* firstArgument)
{  
        bool done = false;
        if (!firstArgument || (strlen(firstArgument) == 1 && *firstArgument == '?') )
        {
                // print out summary information on blank or '?'
                tcout << summaryMap[command] << std::endl;
                done = true;
        }        
        return done;   
}

void 
XSGlobal::registerNativeFitMethods()
{
        FitMethod::registerMethod("leven",new LevMarq("10 0.01"));        
//        FitMethod::registerMethod("anneal",new Anneal);
        summaryMap["method"] =  refreshFitMethodNames();
}

void XSGlobal::initializeGUILibraries(Tcl_Interp* xsInterp)
{
        int status(TCL_OK);
// initialize Tk. If the initialization succeeds, defind the Tk package to Tcl. 
// If it fails, generally speaking this is because the user does not have access
// to and X display, either because of permissions, environment (no DISPLAY variable)
// or because they are not running under X. In this case, simply continue silently
// and do not initialize Tk. 

#ifdef GUI_SUPPORT
        if ( (status = Tk_Init(xsInterp)) == TCL_OK) 
        {
                char tk[] = "Tk";
                Tcl_StaticPackage(xsInterp, tk, Tk_Init, Tk_SafeInit); 
        } 

        // initialize Itcl/Itk

        //
        //  This is itkwish, so import all [incr Tcl] commands by
        //  default into the global namespace.  Fix up the autoloader
        //  to do the same. Use TclInitErr exception to make the flow look simple.
        //  In principle we could load Itcl for the non GUI case, but this is not
        //  too necessary, so we load both or neither.

        if (status == TCL_OK) 
        {
		if ( (status = Itk_Init(xsInterp)) == TCL_OK)
		{
               	        char CItk[] = "Itk";
               	        char itkn[] = "::itk::*";
               	        Tcl_StaticPackage(xsInterp, CItk, Itk_Init, (Tcl_PackageInitProc *) NULL);
                        if ((status = 
                                Tcl_Import(xsInterp, Tcl_GetGlobalNamespace(xsInterp),itkn, 1)) == TCL_OK)
                        {
               	                char mkindex[] = 
               	                       "auto_mkindex_parser::slavehook { _%@namespace import -force ::itcl::* ::itk::* }";
               	                if ( (status = Tcl_Eval(xsInterp,mkindex))  == TCL_ERROR)
                                {
                   	                throw TclInitErr("Itk Error: constructing script index itcl/itk");

                                }
                        }
                        else
                        {
                                throw TclInitErr("Itk Error: getting itk namespace");
                        }
                }
                else
                {
 			throw TclInitErr("Itcl Error: cannot initialize Itk");
		}
        }
        else
        {
                throw TclInitErr("Itcl Error: getting itcl namespace");
        }

        if ( (status = Pow_Init(xsInterp)) == TCL_ERROR)
        {
                std::cerr << "Cannot initialize graphical interface.\n";
                std::cerr << "Switching to command line mode." << std::endl;
                globalData->gui(false); 

        }
#endif
    // NB this necessitates "true" as the default
    // setting since the .xspecrc file hasn't been read yet.

}

void 
XSGlobal::registerNativeStatMethods()
{
        using namespace XSContainer;
        Weight* standard = new StandardWeight;
        Weight* churazov = new ChurazovWeight;
        Weight* gehrels = new GehrelsWeight;
        Weight* modWeight = new ModelWeight;
        StatMethod::registerWeightingMethod(standard->name(), standard);     
        StatMethod::registerWeightingMethod(churazov->name(), churazov);     
        StatMethod::registerWeightingMethod(gehrels->name(), gehrels);     
        StatMethod::registerWeightingMethod(modWeight->name(), modWeight);
        StatMethod::registerMethod("chi",new ChiSquare);
        StatMethod::registerMethod("cstat",new Cstat);  
	StatMethod::registerMethod("lstat",new Lstat); 
        //StatMethod::registerMethod("k-s",new EDF<K_S>);
        //StatMethod::registerMethod("c-vm",new EDF<C_vM>);
        //StatMethod::registerMethod("a-d",new EDF<A_D>);
        string doc("set statistic to optimize \n\tSyntax:statistic [ ");
        doc += StatMethod::statNames();
        doc += " ]";
        summaryMap["statistic"] = doc;
        doc =  "set data weighting\n\tSyntax:weight [ "   ; 
        doc += StatMethod::weightNames();
        doc += " ]";
        summaryMap["weight"] = doc;
}

void 
XSGlobal::registerNativePlotCommands(Plot* plot)
{
        std::pair<string,string> listOfPlotCommands(PlotCommandCreator::registerPlotCommands(plot));
        string& primary = listOfPlotCommands.first;
        string& secondary = listOfPlotCommands.second;        
        string doc ("plot data/models/fits etc\nSyntax: \tprimary plot commands:\n");
        size_t fieldWidth(15);
        size_t npcommands(primary.size() / fieldWidth);
        size_t  cnpl ( npcommands/5);
        for (size_t j = 0; j < cnpl; ++j)
        {
                doc +=  "\t";
                doc +=  primary.substr( j*5*fieldWidth, 5*fieldWidth );
                doc +=  "\n";     
        }
        doc +=  "\t";
        doc += primary.substr(cnpl*5*fieldWidth);
        doc += "\n\tsecondary options ";
        doc += "(2-panel data plots - use with data, ldata, counts, lcounts, icounts):\n";

        size_t nscommands(secondary.size() / fieldWidth);
        size_t  cnsl ( nscommands/5);
        for (size_t j = 0; j < cnsl; ++j)
        {
                doc +=  "\t";
                doc +=  secondary.substr( j*5*fieldWidth, 5*fieldWidth );
                doc +=  "\n";     
        }
        doc +=  "\t";
        doc += secondary.substr(cnsl*5*fieldWidth);
        doc += "\n";
        summaryMap["plot"] = doc;

}

void
XSGlobal::registerNativeTables()
{
    TableComponent::registerTableFormat(new OGIPTable(""));
}

void
XSGlobal::registerNativeLineLists()
{
   LineList::registerLineList("apec", new Apec);
   LineList::registerLineList("bearden", new Bearden);
   LineList::registerLineList("mekal", new Mekal);
}

void 
XSGlobal::startUp()
{


        using namespace XSContainer;

        string title, buildDate;
	XSutility::XSVersionString(title, buildDate);
        tcout << '\n' << "\t\t" << title << std::endl; 
	tcout << '\t' << buildDate << '\n' << std::endl;            

        DataSet::xspecVersion(title);


        // now we can process the other initial settings. 

        globalData->processSettings();


	if (!globalData->gui())
	{
	        plot = Plot::Instance(globalData->defaultGraph(),globalData->managerDir());
	}
	else
	{
	   // This will throw a RedAlert for the time being.
	        plot = Plot::Instance(" ",globalData->managerDir());
	}

        string plotDevice (globalData->settings("PLOTDEVICE"));
	try
	{
        	if ( plotDevice.length())
        	{
                	plot->graph()->setDevice(plotDevice);
        	}  
	}
	catch ( ... )
	{
		// do nothing, the exception will have  printed a
		// message
	}

        //register plot commands. Must happen after the plot control object 
        // is instantiated

        registerNativePlotCommands(plot);


}

void
XSGlobal::cleanUp()
{
   LineList::clearLineLists();
   TableComponent::clearTableFormats();
   FitMethod::clearMethods();
   StatMethod::clearWeightingMethods();
   StatMethod::clearMethods();
   clearFunctionMap();
   delete Help::helpTree();

   // in case the order is important this should be right.
   delete SignalHandler::instance();
   delete XSContainer::plot;
   delete XSContainer::fit;
   delete XSContainer::models;
   delete XSContainer::datasets;
   delete XSContainer::responses;

   system(string("rm -f " + globalData->undoFile()).c_str());

   std::ostringstream cmd;

   string autosaveFile = globalData->autoSaveFile();

   struct stat info;

   stat(autosaveFile.c_str(), &info);

   if(info.st_size)
       cmd << "mv -f " << autosaveFile << ' ' 
	   << autosaveFile.substr(0, autosaveFile.rfind("xautosav_") + 8) << ".xcm";
   else
       cmd << "rm -f " << autosaveFile;

   system((cmd.str()).c_str());
}

void 
XSGlobal::createCommandMap()
{
        string doc("");
        summaryMap["?"] = 
                "print summary of all commands or single command\n\tSyntax: ?{<commandName>}";

        doc =  "reset normalization used in correcting the background\n";
        doc += "\tSyntax: cornorm <range1> <range2>...  norm";
        doc += " (sets default range to last range entered)";
        summaryMap["cornorm"] = doc; 

        doc = "set cosmological parameters\n\tSyntax: cosmo [H0 , q0, lambda0]";
        doc += " (if no arguments are entered, current setting is printed)";
        summaryMap["cosmo"] = doc;

        summaryMap["cpd"] 
                = "change plotting device\n\tSyntax: cpd (device name)";        

        summaryMap["data"] 
                = "input data files\n\tSyntax: data <none> | [n:m] <filename{range}> - (many options: see help for full syntax)";

        summaryMap["fakeit"] 
                = "create simulated observations from model\n\tSyntax: fakeit <none> | <spectrum number> [ background files] [correction files]";

        summaryMap["addcomp"] = 
                "add component to a model\n\tSyntax: addcomp [Model Name:]<new component number> componentName";

        summaryMap["delcomp"] 
                = "delete component from model\n\tSyntax: delcomp [Model Name:]component number";

        doc = "find confidence region\n\tSyntax: ";
        doc += "error [[stopat <ntrial> <tol>] [maximum <redchi>] [<delta fit stat>] [<model param range>...]]"; 
        summaryMap["error"]  = doc;

        doc = "find confidence region\n\tSyntax: ";
        doc += "uncertain [stopat ntrial ntol] [maximum redchi] [delta] param range";
        summaryMap["uncertain"]  = doc;

        doc =  "minimize statistic using current fitting algorithm ";
        doc += "\n\tSyntax: fit [#trials] [critical delta] <method-specific settings>";

        summaryMap["fit"] = doc;

        summaryMap["flux"] 
                = "calculate the model flux\n\tSyntax: flux lowEnergy highEnergy [err #trials confidence level | noerr] ";

        doc = "fix a fit parameter\n\tSyntax: freeze <modelName>:{m1-n1,m2-n2,...},";
        doc += " for integer parameter ranges and for optional model names";

        summaryMap["freeze"] = doc;

	doc = "enter an interactive PLT plot command\n\tSyntax: iplot <?|[plot command]| >\n";
        doc += "\t(iplot with no arguments acts like \"iplot data\")";
	summaryMap["iplot"] = doc;

        doc = "Prepare a package containing locally written models for use with XSPEC";
        doc += "\n\t Syntax: initpackage <packageName> <model init file> [<directory>]"; 
        summaryMap["initpackage"] = doc;

        summaryMap["log"] = 
                "open/close output file\n\t Syntax: log (<filename>|none)";

        doc = "load a package containing locally written models into XSPEC";
        doc += "\n\t Syntax: lmod <packageName> [ <directory> ]";

        summaryMap["lmod"] = doc;

        doc = "calculate the model luminosity\n\tSyntax: ";
        doc += " flux lowEnergy highEnergy [err #trials confidence level | noerr] redshift ";
        summaryMap["lumin"] = doc;



        summaryMap["model"] = 
                "define a model \n\t Syntax: model {<name>:<source>} <model string>";

        doc =  "change model parameters for specified or unnamed model\n";
        doc += "\tSyntax: newpar {modelName:}<parameter range> ";
        doc += "value {<delta> <lower> <bot> <top> <higher>}";        
        summaryMap["newpar"] = doc;

        doc = "change XSPEC querying (ask questions or use supplied answer)";
        doc += "\n\tSyntax: query ( on | yes | no ) ";

        summaryMap["query"] = doc;

        summaryMap["time"] 
                = "display elapsed time since last call and beginning of program run\n\tSyntax: time";

        doc =  "modify plot / change plot device etc\n\tSyntax: setplot ";
        doc += "add|channel|command|delete|device|energy|group|";
        doc += "id|list|noadd|noid|rebin|wave|xlog|ylog";

        summaryMap["setplot"]  = doc;

        doc = "Perform fits while stepping the value of parameters through given range(s)";
        doc += "\n\tSyntax: steppar  <step spec> [ <step spec> ... ]";
        doc += "\n\t      where  <step spec> ::= [log|nolog] [current|best] <param index> <low> <high> <# steps>";
        summaryMap["steppar"]  = doc;


        summaryMap["systematic"] = 
                "set systematic error proportional to model\n\tSyntax: systematic <real value>";

        doc = "allow a previously fixed fit parameter to vary\n\t";
        doc += "Syntax: thaw {m1-n1,m2-n2,...}, for integer parameter ranges";        
        summaryMap["thaw"] = doc;

        doc     =  "unlink parameter(s) previously tied to other parameters\n\t";
        doc     += "Syntax: untie [<model name>:]{range}" ;        
        summaryMap["untie"]  = doc; 


        summaryMap["comsum"] = 
                "print this summary\n\tSyntax: comsum {<commandName>}";
        summaryMap["exit"] = 
                "end the session (also Ctrl-D)\n\tSyntax: exit";


        doc = "Perform a Monte Carlo calculation of the goodness-of-fit\n\t";
        doc += "Syntax: goodness [# realizations] <sim|nosim>";
        summaryMap["goodness"] = doc;

        doc = "List possible lines in the specified energy range\n\t";
        doc += "Syntax: identify <energy> <delta_energy> <redshift> <line_list>";
        summaryMap["identify"] = doc;

        doc = "ignore PHA channels\n\tSyntax: ignore <bad>";
        doc += "\n\tignore s1-<s2|*|**>:n1-n2 ..., integer spectrum number and channel range";
        doc += "\n\tignore s1-<s2|*|**>:r1-r2 ..., integer spectrum number and real energy range";
        doc += "\n\t (multiple ranges are allowed)";
        summaryMap["ignore"]  = doc;

        doc = "notice PHA channels\n\tSyntax: notice all\n";
        doc += "notice <spectrum number range>:<channel range> ...";
        doc += "\n\tnotice <spectrum number range>:<energy range> ...";
        doc += "\n\t (multiple ranges are allowed)";
        summaryMap["notice"] = doc;

	doc = "reset response files\n\tSyntax: response none\n";
	doc += "response [source number:spectrum number] <filename>\n";
	doc += "(many options: see help for full syntax)\n";
        summaryMap["response"] = doc;

	doc = "reset auxiliary response file\n\tSyntax: arf none\n";
	doc += "arf [source number:spectrum number] <filename{range}>\n";
	doc += "(many options: see help for full syntax)\n";
        summaryMap["arf"] = doc;

        doc = "reset background files\n\tSyntax: backgrnd none\n";
	doc += "backgrnd [spectrum number] <filename{range}>\n";
	doc += "(many options: see help for full syntax)\n";
        summaryMap["backgrnd"] = doc; 

        doc = "reset correction files\n\tSyntax: corfile none\n";
	doc += "corfile [spectrum number] <filename{range}>\n";
	doc += "(many options: see help for full syntax)\n";
        summaryMap["corfile"] = doc; 

	doc = "Create a diagonal response.\n\tSyntax: diagrsp\n";
        summaryMap["diagrsp"] = doc;

	doc = "create 'dummy' response\n";
	doc += "\tSyntax: [<Low Energy> [<High Energy> [<# of ranges> [<log or linear> [<channel offset> [<channel width>]]]]]]\n";
        summaryMap["dummyrsp"] = doc;

        doc = "calculate equivalent width\n\tSyntax: eqwidth [[RANGE <frac range>] ";
        doc += "<model component number>]... [err <number> <level> | noerr]";
        summaryMap["eqwidth"] = doc;

        doc = "  Choose from the following XSET options:\n";
        doc += " abund   cosmo   delta   mdatadir   method   seed   statistic   weight   xsect\n";
        doc += "  Model string parameters:\n";
        summaryMap["xset"] = doc;

        summaryMap["chatter"] 
                = "set the chatter level\n\tSyntax: chatter <console> {<log>}";
        summaryMap["ftest"] 
                = "perform f-test\n\tSyntax: ftest chi2 dof2 chi1 dof1 (where chi2 < chi1)\n";        
        summaryMap["lmod"]
                = "load a package containing one or more analytical models\n\tSyntax: localmodel packageName <dir>";
        summaryMap["localmodel"]
                = "load a package containing one or more analytical models\n\tSyntax: localmodel packageName <dir>";
        summaryMap["quit"] = 
                "end the session (also Ctrl-D)\n\tSyntax: quit";
        summaryMap["renorm"] 
                = "adjust the model norms\n\tSyntax: renorm  [ |?|none|prefit|auto ]";
        summaryMap["tclout"] 
                = "write values to tcl variable $xspec_tclout\n\tSyntax: tclout {many options} ";        
        summaryMap["version"]
                = "print version string\n\tSyntax: version (no arguments)";
        summaryMap["script"] = 
                "open/close session script\n\t Syntax: script [filename | none ]";



        summaryMap["editmod"] = 
                "add, replace or delete a component in a model\n\tSyntax: editmod [modelName:]<model expression>";

        summaryMap["help"] 
                = "Invoke XSPEC help system\n\tSyntax: help  [command] [subcommand]";


        summaryMap["save"] 
                = "save current state to a file\n\tSyntax: save <model|files|all> [filename]";

        summaryMap["show"] 
                = "display information about current session state\n\tSyntax: show";

        summaryMap["autosave"] 
                = "set frequency of auto-saving.\n\tSyntax: autosave <off | n>";

        summaryMap["gain"] 
                = "change response gain for a dataset\n\tSyntax: gain <fit|nofit> or: gain <spectrum #> <slope> <intercept>";

        doc = "Run a Monte Carlo Markov Chain.\n\tSyntax: ";
        doc += "[burn <burn-length>|clear|info|length|load|rand|recalc|run|stat|unload]";        
        summaryMap["chain"] = doc;

        doc = "(Bayes only) marginalize over uninteresting parameters.";
        doc += "\n\tSyntax: margin  <step spec> [ <step spec> ... ]";
        doc += "\n\t    where    <step spec> ::= [log|nolog] <param index> <low> <high> <# steps>";
        summaryMap["margin"] = doc;

        doc = "set up Bayesian inference\n\tSyntax: ";  
        doc += "bayes off|on|cons| <mod param #> <prior type> <hyperparameters>";     
        summaryMap["bayes"] = doc;

        doc = "Define an energy array to be used for model calculations in place of";
        doc += "\n\t\tresponse energy arrays.";
        doc += "\n\tSyntax: energies [reset | scale <factor> | <asciiFileName> |";
        doc += "\n\t\t<first range spec> [<additional range spec> ...]]";
        doc += "\n\twhere <first range spec>=::<lowEng> <highEng> <nEngs> log|lin";
        doc += "\n\tand <additional range spec>=::<highEng> <nEngs> log|lin";
        summaryMap["energies"] = doc;

        doc = "Run the MINUIT improve command which attempts to find a new";
        doc += "\n\tlocal minimum in the vicinity of the current minimum.";
        doc += "\n\tThis gives some global minimization capability.";
        doc += "\n\tSyntax: improve";        
        summaryMap["improve"] = doc;

        commandMap["?"] = &xsComsum;
        commandMap["abund"] = &xsAbund;
        commandMap["addcomp"] = &xsAddcomp;
        commandMap["arf"] = &xsArf;
        commandMap["autosave"] = &xsAutosave;
        commandMap["backgrnd"] = &xsBackgrnd;
        commandMap["bayes"] = &xsBayes;
        commandMap["chain"] = &xsChain;
        commandMap["comsum"] = &xsComsum; // broken 11/29/01 - typing in comsum gives
                                          // nonsense.
        commandMap["corfile"] = &xsCorfile;
        commandMap["cornorm"] = &xsCornorm;
        commandMap["cosmo"] = &xsCosmo;
        commandMap["cpd"] = &xsCpd;
        commandMap["chatter"] = &xsChatter;
        commandMap["data"] = &xsData;
        commandMap["delcomp"] = &xsDelcomp;
        commandMap["diagrsp"] = &xsDiagrsp;
        commandMap["dummyrsp"] = &xsDummyrsp;
        commandMap["editmod"] = &xsEditmod;
        commandMap["energies"] = &xsEnergies;
        commandMap["eqwidth"] = &xsEqwidth;
        commandMap["error"] = &xsError;
        commandMap["uncertain"] = &xsError; 
        commandMap["exit"] = &xsExit;
        commandMap["fakeit"] = &xsFakeit;
        commandMap["fit"] = &xsFit;
        commandMap["flux"] = &xsFlux;
        commandMap["lumin"] = &xsFlux;
        commandMap["freeze"] = &xsFreeze;
        commandMap["ftest"] = &xsFtest;
        commandMap["gain"] = &xsGain;
        commandMap["goodness"] = &xsGoodness;
        commandMap["help"] = &xsHelp;
        commandMap["identify"] = &xsIdentify;        
        commandMap["ignore"] = &xsIgnore;
        commandMap["improve"] = &xsImprove;
        commandMap["initpackage"] = &xsInitpackage;
	commandMap["iplot"] = &xsIplot;
        commandMap["log"] = &xsLog;
        commandMap["lmod"] = &xsLmod;
        commandMap["margin"] = &xsMargin;
        commandMap["method"] = &xsMethod;
        commandMap["model"] = &xsModel;
        commandMap["newpar"] = &xsNewpar;
        commandMap["notice"] = &xsNotice;
        commandMap["plot"] = &xsPlot;
        commandMap["query"] = &xsQuery;
        commandMap["quit"] = &xsExit;
        commandMap["renorm"] = &xsRenorm;
        commandMap["response"] = &xsResponse;
        commandMap["script"] = &xsScript;
        commandMap["setplot"] = &xsSetplot;
        commandMap["save"] = &xsSave;        
        commandMap["show"] = &xsShow;
        commandMap["statistic"] = &xsStatistic;
        commandMap["steppar"] = &xsSteppar;
        commandMap["systematic"] = &xsSystematic;
        commandMap["thaw"] = &xsThaw;
        commandMap["tclout"] = &xsTclout;
        commandMap["time"] = &xsTime;
	commandMap["undo"] = &xsUndo;
        commandMap["untie"] = &xsUntie;
        commandMap["version"] = &xsVersion;
        commandMap["weight"] = &xsWeight;
        commandMap["xsect"] = &xsXsect;
        commandMap["xset"] = &xsXset;

        if ( !XSGlobal::globalData->gui()) 
        {       
                // internal, not documented.
                commandMap["unknown"] = &xsUnknown; 
        }
        else 
        {
                // internal, not documented: used by GUI
                // which redefines unknown by itself.
                // XSunknown is the default handler for
                // unknown commands caught this way.
                commandMap["XSunknown"] = &xsUnknown; 
        }


        commandMap["genetic"] = &xsGenetic;
        commandMap["mdefine"] = &xsMdefine;
        commandMap["recornrm"] = &xsRecornrm;       
        commandMap["thleqw"] = &xsThleqw;

        summaryMap["genetic"] 
                = "set genetic algorithm parameters.\n\tSyntax: Not Implemented";               
        summaryMap["mdefine"] 
                = "define an algebraic model component\n\tSyntax: Not Implemented";
        summaryMap["recornrm"] 
                = "adjust correction norms\n\tSyntax: Not Implemented";
        summaryMap["thleqw"] 
                = "fluorescent line equivalent width\n\tSyntax: Not Implemented";


}        

void XSGlobal::registerFunctionUtility ()
{

        FunctionUtility::managerPath(globalData->managerDir());        
        const string HEADAS(getenv("HEADAS"));
        string dataPath (HEADAS + "/../spectral/modelIonData/");      
        FunctionUtility::modelDataPath(dataPath);
        string crossSections;
        string abundances;
        FunctionUtility::readInitializers(crossSections,abundances);
        string xSectDoc("Change/Report the photoionization cross sections\n\t");
        xSectDoc += "Syntax: xsect " + crossSections;
        summaryMap["xsect"] = xSectDoc;
        string abundDoc("Change/Report the solar abundance table in use\n\t");
        abundDoc += "Syntax: abund " + abundances +"\n\t";
        abundDoc += "        abund <name> <file> \n\t"; 
        abundDoc += "        (define & load new vector <name> from data in <file>)";
        summaryMap["abund"] = abundDoc;



}

string XSGlobal::refreshFitMethodNames()
{                       
        string fitMethodDoc("set fitting algorithm");     
        fitMethodDoc += "\n\tSyntax:method [ ";
        fitMethodDoc += FitMethod::fitNames();
        fitMethodDoc += "]  <initial data for method>" ; 
        return fitMethodDoc;               
}

void XSGlobal::commonPlotHandler(const StringArray& args, bool isInteractive)
{
   using XSContainer::plot;
   StringArray plotCmd;

   //lastCommand defaults to "data" in Plot constructor
   plotCmd = (args[0] == "" ? plot->lastCommand() : args);

   const size_t N(plotCmd.size());
   string value(""), arg("");  

   if (N > 0)
   {
      arg = XSutility::lowerCase(plotCmd[0]);
      size_t commaInArg = arg.find_first_of(',');
      if ( commaInArg != XSparse::NOTFOUND()) 
      {
              // the "+1" parses off the first comma which is just a separator
              // between the command and the parameters.
              value = arg.substr(commaInArg+1);
              arg = arg.substr(0,commaInArg);
      }
      else value = Model::DEFAULT();
   }

   if ( N > 1)
   {
        Real dummy;
	size_t dummy2;
        bool isModel = (string("model").find(arg) == 0);

	string tmp, sub = plotCmd[1].substr(0, plotCmd[1].find(','));

        if ( !isModel && !XSutility::isReal(sub, dummy) && 
	     !XSparse::stringIntPair(sub, tmp, dummy2) && plotCmd[1][0] != ',')
        {
	    arg += " ";
	    arg += XSutility::lowerCase(sub);
        }
        // send this and all subsequent arguments to the value string.
        //  "value" string will be ignored by commands that don't use it
        // (all the data commands).
        // commands like "plot model <model name>" will use this.
        // commands like plot contour that are followed by real arguments must
        // deal with by the command.
        value = plotCmd[1];
        for (size_t j = 2; j < N; ++j)
        {
                value += " ";
                value += plotCmd[j];
        }              
   }

   PlotCommand* pC = PlotCommandCreator::commands(XSutility::lowerCase(arg));
   if (!pC)
   {
      throw YellowAlert("Plot subcommand does not exist or has not yet been implemented.\n");
   }
   else
   {
       //save last command
       plot->lastCommand(plotCmd);
      try
      {
	 plot->graph()->setInteractive(isInteractive);
//         std::cerr << value << '\n';
	 pC->execute(value);
      }
      catch (YellowAlert &)
      {
        // If exception propagates after PlotGroups are created,
        // but before command has a chance to call plot->makeTheGraph
        // (which deletes them), then needs to be cleaned up here.
         plot->resetPlotGroups();
	 throw;
      }
   }
}

void XSGlobal::saveAll(const string& outFileName)
{
    try
    {
	std::ofstream outFile(outFileName.c_str());
	// note that xspec11 also prints out gain
	// information - since the gain command is
	// not yet implemented in xspec12 this 
	// is not done here but needs to be addressed.
	// BD 11/16/03
	XSContainer::datasets->saveData(outFile);
	saveModel(outFile);
    }
    catch ( std::exception& )
    {
	tcerr << " File cannot be opened for writing\n"; 
	// disable autosave to forbid further messages.  
	globalData->autoSaveFrequency(XSparse::NOTFOUND());
    }
}

void XSGlobal::saveModel(std::ofstream& outStream)
{
    using namespace std;
    FitMethod* fitMethod  (XSContainer::fit->fitMethod());
    StatMethod* statMethod (XSContainer::fit->statMethod());

    outStream << "statistic "  << statMethod->name() << endl;
    outStream << "method "  << fitMethod->selectedSubMethod() << ' ' 
	    << fitMethod->settingString()   << endl;
    if(FunctionUtility::abundChanged())
        outStream << "cd " << FunctionUtility::abundPath() << "\nabund file " 
	          << FunctionUtility::abundanceFile() << endl;
    else
        outStream << "abund " << FunctionUtility::ABUND() << endl;
    outStream << "xsect " << FunctionUtility::XSECT() << endl;
    outStream << "cosmo " << FunctionUtility::getH0() << " "
        << FunctionUtility::getq0() << " " << FunctionUtility::getlambda0()
        << endl;
    outStream << "xset delta " 
        << XSContainer::models->proportionalDelta() << endl;
    const map<string,string>& xsetStrings = 
            FunctionUtility::modelStringDataBase();
    map<string,string>::const_iterator itString = xsetStrings.begin();
    map<string,string>::const_iterator itStringEnd = xsetStrings.end();
    while (itString != itStringEnd)
    {
       outStream << "xset " << itString->first << "  " 
            << itString->second << endl;
       ++itString;
    }

    XSContainer::models->saveData(outStream);
}

void XSGlobal::registerNativeRandomizingStrategies(Fit* fit)
{
   std::auto_ptr<RandomizerBase> rand(new Randomizer<Fit,RandomizerGaussDist>());
   // If the following call doesn't throw, relinquish ownership of 
   // rand to the fit container.
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   RandomizerBase* tmp = rand.release();
   fit->chainManager()->setChainProposal(tmp->name());
   rand.reset(new Randomizer<ChainManager,RandomizerGaussDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();
   rand.reset(new Randomizer<RandomizerCmdLine,RandomizerGaussDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();
   rand.reset(new Randomizer<RandomizerCovarFile,RandomizerGaussDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();

   rand.reset(new Randomizer<Fit,RandomizerCauchyDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();
   rand.reset(new Randomizer<ChainManager,RandomizerCauchyDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();
   rand.reset(new Randomizer<RandomizerCmdLine,RandomizerCauchyDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();
   rand.reset(new Randomizer<RandomizerCovarFile,RandomizerCauchyDist>());
   fit->registerRandomizingStrategy(rand->name(), rand.get());
   rand.release();


   // We want to distinguish these native classes from any the
   // user may add in at a later time.  So fill in Fit's
   // nativeRandomizerNames container with these names.
   // (This can only be done once.)
   fit->initNativeRandomizerNames();
}
